import unittest

from geomalgo import Point2D, Polygon2D, Triangle2D, Segment2D

class TestSquareWinding(unittest.TestCase):

    """
    1  D-------C
       |       |
       |       |
       |       |
    0  A-------B
       0        1
    """

    A = Point2D(0, 0)
    B = Point2D(1, 0)
    C = Point2D(1, 1)
    D = Point2D(0, 1)

    polygon2d = Polygon2D.from_points2d([A,B,C,D])

    def test_point_is_inside(self):
        P = Point2D(0.5, 0.5)
        self.assertTrue(self.polygon2d.includes_point(P))

    def test_point_is_outside(self):
        P = Point2D(1.5, 0.5)
        self.assertFalse(self.polygon2d.includes_point(P))


class TriangleWinding(unittest.TestCase):

    """
    Base class for Polygon2D and Triange2D.

    Triangle2D use the same algorithm that Polygon2d, but specialized
    for optimization.

    """

    def assert_inside(self, A, B, C, P, edge_width=0.):
        triangle = Triangle2D(A, B, C)
        #polygon = Polygon2D.from_points2d([A,B,C])

        self.assertTrue(triangle.includes_point(P, edge_width=edge_width))
        #self.assertTrue(polygon.includes_point(P) )

    def assert_outside(self, A, B, C, P):
        triangle = Triangle2D(A, B, C)
        polygon = Polygon2D.from_points2d([A,B,C])

        self.assertFalse(triangle.includes_point(P))
        self.assertFalse(polygon.includes_point(P) )

    def test_inside(self):
        A = Point2D(0, 0)
        B = Point2D(1, 0)
        C = Point2D(0, 1)
        P = Point2D(0.25, 0.25)
        self.assert_inside(A, B, C, P)

    def test_outside(self):
        A = Point2D(0, 0)
        B = Point2D(1, 0)
        C = Point2D(0, 1)
        P = Point2D(1., 1.)
        self.assert_outside(A, B, C, P)

    def test_inside_on_edge(self):
        # Example taken from:
        # https://totologic.blogspot.fr/2014/01/accurate-point-in-triangle-test.html
        A = Point2D(1/10, 1/9)
        B = Point2D(100/8, 100/3)
        C = Point2D(100/4, 100/9)
        D = Point2D(-100/8, 100/6)
        AB = Segment2D(A, C)
        P = A + (B-A)*(3/7)
        self.assert_inside(A, B, C, P, edge_width=1.E-03)

if __name__ == '__main__':
    unittest.main()



from .triangulation2d cimport Triangulation2D
from ..grid2d cimport Grid2D

cdef class TriangulationLocator:

    cdef public:
        Triangulation2D TG
        Grid2D grid

        double edge_width
        double edge_width_square

        int[:] celltri
        int[:] celltri_idx

    cpdef int[:] search_points(TriangulationLocator self,
                               double[:] xpoints, double[:] ypoints,
                               int[:] triangles=*)



from .plane import Plane
from .point3d import Point3D
from .segment3d import Segment3D
from .triangle3d import Triangle3D
from .vector3d import Vector3D

__all__ = [
    'Plane', 'Point3D', 'Segment3D', 'Triangle3D', 'Vector3D',
]



import unittest
from math import sqrt

from numpy.testing import assert_allclose

import geomalgo as ga


STEP = ga.data.step


class TestUtil(unittest.TestCase):

    def test_centers(self):
        TG = ga.Triangulation2D(STEP.x, STEP.y, STEP.trivtx)

        xcenter, ycenter = ga.triangulation.compute_centers(TG)

        a = 1/3
        b = 2/3
        assert_allclose(xcenter, [a, 1.5, b, 2, a, b])
        assert_allclose(ycenter, [10+a, 10+a, 10+b, 10+b, 11+a, 11+b])

    def test_signed_area(self):
        # make some triangle clockwise to have negative signed area
        trivtx = STEP.trivtx.copy()
        trivtx[0, 0], trivtx[0, 1] = trivtx[0, 1], trivtx[0, 0]
        trivtx[1, 1], trivtx[1, 2] = trivtx[1, 2], trivtx[1, 1]

        TG = ga.Triangulation2D(STEP.x, STEP.y, trivtx)

        signed_area = ga.triangulation.compute_signed_area(TG)

        assert_allclose(signed_area, [-0.5, -0.75, 0.5, 0.75, 0.5, 0.5])

    def test_compute_bounding_box(self):
        TG = ga.Triangulation2D(STEP.x, STEP.y, STEP.trivtx)
        bb = ga.triangulation.compute_bounding_box(TG)

        self.assertEqual(bb.xmin, 0)
        self.assertEqual(bb.xmax, 2.5)

        self.assertEqual(bb.ymin, 10)
        self.assertEqual(bb.ymax, 12)

    def test_compute_edge_min_max(self):
        TG = ga.Triangulation2D(STEP.x, STEP.y, STEP.trivtx)
        edge_min, edge_max = ga.triangulation.compute_edge_min_max(TG)

        self.assertAlmostEqual(edge_min, 1)
        self.assertAlmostEqual(edge_max, sqrt(3.25))


if __name__ == '__main__':
    unittest.main()



cpdef inline double symetric1d(double center, double x):
    """
    +------+------+
    x    center   symetric
    """
    # center = (x + symetric) / 2
    # so
    # symetric = 2*center - x
    return 2*center - x




import unittest
from math import sqrt

from geomalgo import Vector2D

class TestVector(unittest.TestCase):

    def test_dot_product(self):
        A = Vector2D(1,-2)
        B = Vector2D(3,4)
        x = A.dot(B)
        self.assertEqual(x, -5)

    def test_norm(self):
        A = Vector2D(1, 0)
        self.assertEqual(A.norm, 1.)

        A = Vector2D(1, 2)
        self.assertAlmostEqual(A.norm, sqrt(5.))

    def test_subtract_vector(self):
        A = Vector2D(1,2)
        B = Vector2D(4,6)
        V = B - A
        self.assertEqual(V.x, 3)
        self.assertEqual(V.y, 4)

    def test_add_vector(self):
        A = Vector2D(1,2)
        B = Vector2D(4,6)
        V = A + B
        self.assertEqual(V.x, 5)
        self.assertEqual(V.y, 8)

    def test_mult_scalar(self):
        u = Vector2D(1,-2)

        w = u * 3
        self.assertEqual(w.x,  3)
        self.assertEqual(w.y, -6)

    def test_cross_product(self):
        A = Vector2D(1,-2)
        B = Vector2D(3,4)
        x = A ^ B
        self.assertEqual(x, 10)

    def test_normalize(self):
        A = Vector2D(3, 4)
        A.normalize()
        self.assertAlmostEqual(A.x, 0.6)
        self.assertAlmostEqual(A.y, 0.8)
        self.assertAlmostEqual(A.norm, 1.)

class TestNormal(unittest.TestCase):

    def test_horizontal(self):
        """
        vec
        ---------->  |
                     | normal
                     v
        """
        V = Vector2D(3, 0)
        N = V.normal
        self.assertAlmostEqual(N.x, 0)
        self.assertAlmostEqual(N.y, -1)

    def test_horizontal_negative(self):
        """
        ^ normal
        |          vec
        |  <----------
        """
        V = Vector2D(-0.2, 0)
        N = V.normal
        self.assertAlmostEqual(N.x, 0)
        self.assertAlmostEqual(N.y, 1)

    def test_vertical(self):
        """
        ^ vec
        |         normal
        |  --------->
        """
        V = Vector2D(0, 0.2)
        N = V.normal
        self.assertAlmostEqual(N.x, 1)
        self.assertAlmostEqual(N.y, 0)

    def test_vertical_negative(self):
        """
        normal
        <---------  |
                    | vec
                    v
        """
        V = Vector2D(0, -0.2)
        N = V.normal
        self.assertAlmostEqual(N.x, -1)
        self.assertAlmostEqual(N.y,  0)

    def test_oblique(self):
        """
                   _
                   /| vec
                /
            / \
        /      \
               _\| normal
        """
        V = Vector2D(3, 4)
        N = V.normal
        self.assertAlmostEqual(N.x,  0.8)
        self.assertAlmostEqual(N.y, -0.6)


if __name__ == '__main__':
    unittest.main()



import unittest

from geomalgo import Point3D, Segment3D

class TestTriangle(unittest.TestCase):

    def test_create_segment(self):
        A = Point3D(1,2,3)
        B = Point3D(6,5,3)

        segment = Segment3D(A,B)
        self.assertEqual(segment.B.y, 5)

        segment.B.y = 3
        self.assertEqual(segment.B.y, 3)

        segment.B = Point3D(-1, -2, -3)
        self.assertEqual(segment.B.y, -2)

if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

import geomalgo as ga

HOLE = ga.data.hole

class TestTriangulationInterpolator(unittest.TestCase):

    def test_normal(self):

        def linear_function(x,y):
            """Use a linear function, so linear interpolation will be exact"""
            return -4*np.asarray(x) + 7*np.asarray(y) - 16

        # Create triangulation and a locator.
        TG = ga.Triangulation2D(HOLE.x, HOLE.y, HOLE.trivtx)
        locator = ga.TriangulationLocator(TG)

        # Create interpolator to triangle centers.
        interpolator = ga.TriangulationInterpolator(TG, locator, TG.NT)
        xcenter, ycenter = ga.triangulation.compute_centers(TG)
        nout = interpolator.set_points(xcenter, ycenter)
        self.assertEqual(nout, 0)

        # Compute data on mesh vertices, and expected data on triangle
        # centers.
        vertdata = linear_function(TG.x, TG.y)
        expected_pointdata = linear_function(xcenter, ycenter)

        # Interpolate data.
        pointdata = np.empty(TG.NT, dtype='d')
        interpolator.interpolate(vertdata, pointdata)

        # Check interpolated data.
        self.assertEqual(nout, 0)
        np.testing.assert_allclose(pointdata, expected_pointdata)

if __name__ == '__main__':
    unittest.main()




from .triangulation2d cimport Triangulation2D
from .locator cimport TriangulationLocator

cdef class TriangulationInterpolator:
    """
    Interpolate data from triangle vertices to points.

    """

    cdef public:
        TriangulationLocator locator

        Triangulation2D TG

        int NP

        double[:,:] gradx
        double[:,:] grady
        double[:,:] det

        # Point localization.
        int[:] triangles

        # Interpolation factors
        double[:,:] factors

    cpdef int set_points(TriangulationInterpolator self,
                         double[:] xpoints, double[:] ypoints)

    cpdef void interpolate(self, double[:] tridata, double[:] pointdata)



from .triangulation2d import Triangulation2D
from .edge_to_triangle import EdgeToTriangles
from .boundary_edges import BoundaryEdges
from .intern_edges import InternEdges
from .build_edges import build_edges
from .locator import (
    build_triangle_to_cell, build_cell_to_triangle, TriangulationLocator,
)
from .interpolator import TriangulationInterpolator

__all__ = [
    'Triangulation2D', 'EdgeToTriangles', 'BoundaryEdges', 'InternEdges',
    'build_edges', 'build_triangle_to_cell', 'build_cell_to_triangle',
    'TriangulationLocator', 'TriangulationInterpolator',
]


# ==============
# not in __all__
# ==============


from .util import (
    compute_bounding_box, compute_edge_min_max, compute_centers,
    compute_signed_area, compute_interpolator
)

from .locator import (
    choose_edge_width, choose_bb_grid_distance, build_grid, check_grid
)



import unittest

from geomalgo import (
    Point3D, Triangle3D, Segment3D, intersec3d_triangle_segment
)

class TestIntersection(unittest.TestCase):

    def test_intersection_in_triangle(self):
        """
        C      T       y ^
        |  - /           | z
        |   I   -        |/
        A--/--------B    +-----> x
          S
        """

        A = Point3D(0, 0, 0)
        B = Point3D(1, 0, 0)
        C = Point3D(0, 1, 0)
        triangle = Triangle3D(A,B,C)

        S = Point3D(0.25, 0.25, -1)
        T = Point3D(0.25, 0.25,  1)
        segment = Segment3D(S,T)

        I = intersec3d_triangle_segment(triangle, segment)

        self.assertEqual(I.x, 0.25)
        self.assertEqual(I.y, 0.25)
        self.assertEqual(I.z, 0.)

    def test_intersection_outside_triangle(self):
        A = Point3D(0, 0, 0)
        B = Point3D(1, 0, 0)
        C = Point3D(0, 1, 0)
        triangle = Triangle3D(A,B,C)

        S = Point3D(1., 1., -1)
        T = Point3D(1., 1.,  1)
        segment = Segment3D(S,T)

        with self.assertRaisesRegex(ValueError, 'There is no intersection'):
            intersec3d_triangle_segment(triangle, segment)

    def test_no_intersection(self):
        A = Point3D(0, 0, 0)
        B = Point3D(1, 0, 0)
        C = Point3D(0, 1, 0)
        triangle = Triangle3D(A,B,C)

        S = Point3D(0.25, 0.25, -1)
        T = Point3D(0.25, 0.25, -0.5)
        segment = Segment3D(S,T)

        with self.assertRaisesRegex(ValueError, 'There is no intersection'):
            intersec3d_triangle_segment(triangle, segment)

    def test_triangle_is_a_segment(self):
        A = Point3D(0, 0, 0)
        B = Point3D(0, 0, 0)
        C = Point3D(0, 1, 0)
        triangle = Triangle3D(A,B,C)

        S = Point3D(0.25, 0.25, -1)
        T = Point3D(0.25, 0.25,  1)
        segment = Segment3D(S,T)

        msg = 'Triangle is degenerated \(a segment or point\)'
        with self.assertRaisesRegex(ValueError, msg):
            intersec3d_triangle_segment(triangle, segment)

    def test_triangle_is_a_point(self):
        A = Point3D(0, 0, 0)
        B = Point3D(0, 0, 0)
        C = Point3D(0, 0, 0)
        triangle = Triangle3D(A,B,C)

        S = Point3D(0.25, 0.25, -1)
        T = Point3D(0.25, 0.25,  1)
        segment = Segment3D(S,T)

        msg = 'Triangle is degenerated \(a segment or point\)'
        with self.assertRaisesRegex(ValueError, msg):
            intersec3d_triangle_segment(triangle, segment)

    def test_triangle_and_segment_are_in_same_plane(self):
        A = Point3D(0, 0, 0)
        B = Point3D(1, 0, 0)
        C = Point3D(0, 1, 0)
        triangle = Triangle3D(A,B,C)

        S = Point3D(0., 0.5, 0)
        T = Point3D(0.5, 0., 0)
        segment = Segment3D(S,T)

        msg = 'Triangle and segment are in the same plane'
        with self.assertRaisesRegex(ValueError, msg):
            intersec3d_triangle_segment(triangle, segment)

    def test_triangle_and_segment_are_parallel(self):
        A = Point3D(0, 0, 0)
        B = Point3D(1, 0, 0)
        C = Point3D(0, 1, 0)
        triangle = Triangle3D(A,B,C)

        S = Point3D(0., 0.5, 1)
        T = Point3D(0.5, 0., 1)
        segment = Segment3D(S,T)

        with self.assertRaisesRegex(ValueError, 'There is no intersection'):
            intersec3d_triangle_segment(triangle, segment)

if __name__ == '__main__':
    unittest.main()



from ..base2d.point2d cimport CPoint2D
from ..base2d.segment2d cimport CSegment2D

cdef int intersect_segment2d_segment2d(CSegment2D* S1, CSegment2D* S2,
                                       CPoint2D* I0, CPoint2D* I1,
                                       double coords[4],
                                       double epsilon=*)



import unittest
from math import sqrt

import numpy as np

from geomalgo import Point2D, Line2D

class TestLine2D(unittest.TestCase):

    def test_create_line(self):
        A = Point2D(1,2)
        B = Point2D(3,4)

        # ======================
        # Create line2d.
        # ======================
        line = Line2D(A,B)
        self.assertEqual(line.B.y, 4)

        # ======================
        # Modify B.y
        # ======================
        line.B.y = 5
        self.assertEqual(line.B.y, 5)

        # ======================
        # Modify B
        # ======================
        line.B = Point2D(-1, -2)
        self.assertEqual(line.B.y, -2)


class TestPoint2dDistance(unittest.TestCase):

    def test_horizontal_line(self):
        """
              P

        A-----------B
        """

        A = Point2D(1, 2)
        B = Point2D(3, 2)
        AB = Line2D(A, B)

        P = Point2D(2, 4)
        self.assertEqual(AB.point_distance(P), 2)

        P = Point2D(10, 4)
        self.assertEqual(AB.point_distance(P), 2)

    def test_vertical_line(self):
        """
        B
        |
        |  B
        |
        A
        """

        A = Point2D(1, 2)
        B = Point2D(1, 4)
        AB = Line2D(A, B)

        P = Point2D(3, 3)
        self.assertEqual(AB.point_distance(P), 2)

        P = Point2D(3, 10)
        self.assertEqual(AB.point_distance(P), 2)


    def test_on_horizontal_line(self):
        """
        A-----P-----B
        """

        A = Point2D(1, 2)
        B = Point2D(3, 2)
        AB = Line2D(A, B)

        P = Point2D(2, 2)
        self.assertEqual(AB.point_distance(P), 0)

        P = Point2D(10, 2)
        self.assertEqual(AB.point_distance(P), 0)


if __name__ == '__main__':
    unittest.main()



from .point2d cimport CPoint2D, CVector2D, Point2D, Vector2D, subtract_points2d


# ============================================================================
# Structures
# ============================================================================


cdef struct CSegment2D:
    CPoint2D* A
    CPoint2D* B

cdef CSegment2D* new_segment2d()

cdef void del_segment2d(CSegment2D* csegment2d)

cdef inline void segment2d_set(CSegment2D* AB, CPoint2D* A, CPoint2D* B):
    AB.A = A
    AB.B = B


# ============================================================================
# Computational functions
# ============================================================================


cdef double segment2d_distance_point2d(CSegment2D* AB, CVector2D* u,
                                       CPoint2D* P)

cdef double segment2d_square_distance_point2d(CSegment2D* AB, CVector2D* u,
                                              CPoint2D* P)

cdef double segment2d_where(CPoint2D* A, CVector2D* AB, CPoint2D* P)

cdef void segment2d_middle(CPoint2D* M, CSegment2D* AB)


# ============================================================================
# Python API
# ============================================================================


cdef class Segment2D:

    cdef readonly:
        Vector2D u
        double length

    cdef:
        Point2D A
        Point2D B
        CSegment2D csegment2d

    cdef alloc_new(Segment2D self)

cdef class Segment2DCollection:

    cdef:
        double[:] xa
        double[:] xb

        double[:] ya
        double[:] yb

        int size

    cdef void get(Segment2DCollection self, int segment_index,
                  CSegment2D* segment)

    cdef void set(Segment2DCollection self, int segment_index,
                  CPoint2D* A, CPoint2D * B)



from .boundingbox import BoundingBox
from .line2d import Line2D
from .point2d import Point2D
from .polygon2d import Polygon2D
from .segment2d import Segment2D, Segment2DCollection
from .triangle2d import Triangle2D
from .vector2d import Vector2D

__all__ = [
    'BoundingBox', 'Line2D', 'Point2D', 'Polygon2D', 'Segment2D',
    'Segment2DCollection', 'Triangle2D', 'Vector2D',
]



import unittest

import numpy as np

from geomalgo import (
    Point2D, Grid2D, compute_index, compute_row_col, coord_to_index
)

from geomalgo.data import step


class TestCoordToIndex(unittest.TestCase):

    def test_normal(self):
        """Test index is computed from coordinate"""
        #  -2   -1.5   -1   -0.5    0    0.5    1    1.5    2
        #   |     |     |     |     |     |     |     |     |
        #   |  0  |  1  |  2  |  3  |  4  |  5  |  6  |  7  |
        minval = -2
        delta = 0.5

        self.assertEqual(coord_to_index( 0.75, minval, delta), 5)
        self.assertEqual(coord_to_index(-0.75, minval, delta), 2)


class TestComputeIndex(unittest.TestCase):

    def test_normal(self):
        """Test cell index is computed from cell ix and iy"""
        #  iy
        #     +---+---+---+
        #   1 | 3 | 4 | 5 |
        #     +---+---+---+
        #   0 | 0 | 1 | 2 |
        #     +---+---+---+
        #       0   1   2   ix
        nx = 3

        self.assertEqual(compute_index(nx, 0, 0), 0)
        self.assertEqual(compute_index(nx, 1, 0), 1)
        self.assertEqual(compute_index(nx, 2, 0), 2)

        self.assertEqual(compute_index(nx, 0, 1), 3)
        self.assertEqual(compute_index(nx, 1, 1), 4)
        self.assertEqual(compute_index(nx, 2, 1), 5)


class TestComputeRowCol(unittest.TestCase):

    def test_normal(self):
        """Test cell ix and iy are computed from cell index"""
        #  iy
        #     +---+---+---+
        #   1 | 3 | 4 | 5 |
        #     +---+---+---+
        #   0 | 0 | 1 | 2 |
        #     +---+---+---+
        #       0   1   2   ix

        nx = 3

        self.assertEqual(compute_row_col(0, nx), (0, 0))
        self.assertEqual(compute_row_col(1, nx), (1, 0))
        self.assertEqual(compute_row_col(2, nx), (2, 0))

        self.assertEqual(compute_row_col(3, nx), (0, 1))
        self.assertEqual(compute_row_col(4, nx), (1, 1))
        self.assertEqual(compute_row_col(5, nx), (2, 1))


class TestGrid2D(unittest.TestCase):

    def test_find_cell(self):
        """Test the cell containing a point is found"""
        #     iy
        #  30    +-------+-------+-------+-------+
        #        |     4 |     5 |     6 |     7 |
        #      1 |       |       |   P   |       |
        #        |       |       |       |       |
        #  20    +-------+-------+-------+-------+
        #        |     0 |     1 |     2 |     3 |
        #      0 |   Q   |       |       |       |
        #        |       |       |       |       |
        #  10    +-------+-------+-------+-------+
        #        -1      -0.5    0       0.5     1
        #            0       1       2       3     ix

        grid = Grid2D(xmin=-1, xmax=1.0, nx=4, ymin=10, ymax=30, ny=2)
        P = Point2D(0.25, 25)
        Q = Point2D(-0.75, 15)

        cell = grid.find_cell(P)
        self.assertEqual(cell.ix, 2)
        self.assertEqual(cell.iy, 1)
        self.assertEqual(cell.index, 6)

        cell = grid.find_cell(Q)
        self.assertEqual(cell.ix, 0)
        self.assertEqual(cell.iy, 0)
        self.assertEqual(cell.index, 0)


if __name__ == '__main__':
    unittest.main()



from .point3d cimport CPoint3D, Point3D, CVector3D, Vector3D

cdef struct CPlane:
    CPoint3D* point
    CVector3D* normal

cdef CPlane* new_plane()

cdef void del_plane(CPlane* plane)

cdef void point_projection_to_plane(CPoint3D* point, CPlane* plane,
                                    CPoint3D* projection)

cdef class Plane:

    cdef public:
        Point3D point
        Vector3D normal



cdef class EdgeMap:
    cdef public:
        int NV
        int NE
        int[:] bounds
        int[:] edges
        int[:] location
        int[:] idx

    cdef int search_edge_idx(EdgeMap self, int V0, int V1, bint* found)

    cdef int search_next_boundary_edge(EdgeMap self, int V0, int V1)

    cdef int search_edge_idx_inf(EdgeMap self, int V0, int V1, bint* found)

    cpdef int[:] neighbours(EdgeMap self, int V0)



from libc.math cimport sqrt

cdef struct CVector3D:
    double x
    double y
    double z

cdef CVector3D* new_vector3d()

cdef void del_vector3d(CVector3D* V)

cdef inline void cross_product3d(CVector3D *c, CVector3D *a, CVector3D *b):
    c.x = a.y*b.z - a.z*b.y
    c.y = a.z*b.x - a.x*b.z
    c.z = a.x*b.y - a.y*b.x

cdef inline void subtract_vector3d(CVector3D *c, CVector3D *b, CVector3D *a):
    c.x = b.x - a.x
    c.y = b.y - a.y
    c.z = b.z - a.z

cdef inline double dot_product3d(CVector3D *a, CVector3D *b):
    return a.x*b.x + a.y*b.y + a.z*b.z

cdef inline double compute_norm3d(CVector3D *a):
    return sqrt( a.x*a.x + a.y*a.y + a.z*a.z )

cdef inline void normalize_vector3d(CVector3D *a):
    cdef:
        double norm = compute_norm3d(a)
    a.x /= norm
    a.y /= norm
    a.z /= norm

cdef class Vector3D:
    cdef:
        CVector3D* cvector3d



from .base2d.boundingbox cimport (
    BoundingBox
)

from .base2d.line2d cimport (
    CLine2D, new_line2d, del_line2d, create_line2d
)

from .base2d.point2d cimport (
    CPoint2D, new_point2d, del_point2d, subtract_points2d,
    point2d_plus_vector2d, is_left, point2d_square_distance,
    point2d_distance, point2d_equal, Point2D
)

from .base2d.polygon2d cimport (
    CPolygon2D, new_polygon2d, del_polygon2d, Polygon2D
)

from .base2d.segment2d cimport (
    CSegment2D, new_segment2d, del_segment2d,
    segment2d_square_distance_point2d, segment2d_distance_point2d,
    segment2d_where, segment2d_middle, Segment2D,
    Segment2DCollection, segment2d_set
)

from .base2d.triangle2d cimport (
    CTriangle2D, new_triangle2d, del_triangle2d, triangle2d_includes_point2d,
    triangle2d_on_edges, triangle2d_signed_area, triangle2d_area,
    triangle2d_center, triangle2d_set,
    triangle2d_gradx_grady_det, Triangle2D
)

from .base2d.vector2d cimport (
    CVector2D, new_vector2d, del_vector2d, cross_product2d, subtract_vector2d,
    add_vector2d, vector2d_times_scalar, dot_product2d, compute_norm2d, normalize_vector2d,
    compute_normal2d, Vector2D
)



import unittest

import numpy as np

from geomalgo import Point2D, Polygon2D

class TestPolygon2D(unittest.TestCase):

    def test_create(self):
        """
        2  F----E
           |    |
        1  |    D----C
           |         |
        0  A---------B
           0    1    2
        """
        A = Point2D(0, 0)
        B = Point2D(2, 0)
        C = Point2D(2, 1)
        D = Point2D(1, 1)
        E = Point2D(1, 2)
        F = Point2D(0, 2)

        polygon2d = Polygon2D.from_points2d([A,B,C,D,E,F,A])

        np.testing.assert_equal(polygon2d.x, [0., 2., 2., 1., 1., 0., 0.])
        np.testing.assert_equal(polygon2d.y, [0., 0., 1., 1., 2., 2., 0.])

if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from numpy.testing import assert_equal

from geomalgo import build_edges
from geomalgo.data import step, hole

class TestStep(unittest.TestCase):

    def setUp(self):
        self.intern_edges, self.boundary_edges, _ = build_edges(step.trivtx,
                                                                step.NV)

        self.boundary_edges.add_label(step.boundary_edge_label)
        self.boundary_edges.compute_length(step.x, step.y)
        self.boundary_edges.compute_normal(step.x, step.y)

    def test_internal_edges_vertices(self):
        vert = np.asarray(self.intern_edges.vertices)
        self.assertEqual(vert.shape, (step.NI, 2))
        assert_equal(vert[0], (1,3))
        assert_equal(vert[1], (1,4))
        assert_equal(vert[2], (2,4))
        assert_equal(vert[3], (3,4))
        assert_equal(vert[4], (4,6))

    def test_internal_edges_triangles(self):
        tri = np.asarray(self.intern_edges.triangles)
        self.assertEqual(tri.shape, (step.NI, 2))
        assert_equal(tri[0], (0,2))
        assert_equal(tri[1], (2,1))
        assert_equal(tri[2], (1,3))
        assert_equal(tri[3], (4,2))
        assert_equal(tri[4], (4,5))

    def test_boundary_edges_vertices(self):
        vert = np.asarray(self.boundary_edges.vertices)
        self.assertEqual(vert.shape, (step.NB, 2))
        assert_equal(vert[0], (0,1))
        assert_equal(vert[1], (3,0))
        assert_equal(vert[2], (1,2))
        assert_equal(vert[3], (2,5))
        assert_equal(vert[4], (6,3))
        assert_equal(vert[5], (5,4))
        assert_equal(vert[6], (4,7))
        assert_equal(vert[7], (7,6))

    def test_boundary_edges_triangle(self):
        tri = np.asarray(self.boundary_edges.triangle)
        self.assertEqual(tri.shape, (step.NB,))
        assert_equal(tri[0], 0)
        assert_equal(tri[1], 0)
        assert_equal(tri[2], 1)
        assert_equal(tri[3], 3)
        assert_equal(tri[4], 4)
        assert_equal(tri[5], 3)
        assert_equal(tri[6], 5)
        assert_equal(tri[7], 5)

    def test_boundary_edge_next(self):
        nextbe = np.asarray(self.boundary_edges.next_boundary_edge)
        self.assertEqual(nextbe.shape, (step.NB,))
        self.assertEqual(nextbe[0], 2) # (0,1), (1,2)
        self.assertEqual(nextbe[1], 0) # (3,0), (0,1)
        self.assertEqual(nextbe[2], 3) # (1,2), (2,5)
        self.assertEqual(nextbe[3], 5) # (2,5), (5,4)
        self.assertEqual(nextbe[4], 1) # (6,3), (3,0)
        self.assertEqual(nextbe[5], 6) # (5,4), (4,7)
        self.assertEqual(nextbe[6], 7) # (4,7), (7,6)
        self.assertEqual(nextbe[7], 4) # (7,6), (6,3)

    def test_edge_map(self):
        edge_map = self.intern_edges.edge_map

        bounds = np.asarray(edge_map.bounds)
        edges = np.asarray(edge_map.edges)
        location = np.asarray(edge_map.location)
        idx = np.asarray(edge_map.idx)

        # Indices for edges.
        assert_equal(bounds, [
             0,  2, # 0
             2,  5, # 1
             6,  8, # 2
             9, 11, # 3
            13, 16, # 4
            19, 19, # 5
            21, 22, # 6
            24, 24, # 7
            26,
        ])

        assert_equal(edges, [
                     # V0    bounds  edges

            1, 3,    # 0 =>  0:2  => (0,1), (0,3)
                     # 0 =>  2:2  =>

            3, 2, 4, # 1 =>  2:5  => (1,3), (1,2), (1,4)
            0,       # 1 =>  5:6  => (1,0)

            4, 5,    # 2 =>  6:8  => (2,4), (2,5)
            1,       # 2 =>  8:9  => (2,1)

            4, 6,    # 3 =>  9:11 => (3,4), (3,6)
            1, 0,    # 3 => 11:13 => (3,1), (3,0)

            5, 6, 7, # 4 => 13:16 => (4,5), (4,6), (4,7)
            2, 1, 3, # 4 => 16:19 => (4,2), (4,1), (4,3)

                     # 5 => 19:19 =>
            2, 4,    # 5 => 19:21 => (5,2), (5,4)

            7,       # 6 => 21:22 => (6,7)
            4, 3,    # 6 => 22:24 => (6,4), (6,3)

                     # 7 => 24:24 =>
            4, 6,    # 7 => 24:26 => (7,4), (7,6)
        ])

        # Whether to use idx for boundary_edges or intern_edges
        assert_equal(location, [
                     # V0    bounds  edges

            1, 1,    # 0 =>  0:2  => (0,1), (0,3)
                     # 0 =>  2:2  =>

            0, 1, 0, # 1 =>  2:5  => (1,3), (1,2), (1,4)
            1,       # 1 =>  5:6  => (1,0)

            0, 1,    # 2 =>  6:8  => (2,4), (2,5)
            1,       # 2 =>  8:9  => (2,1)

            0, 1,    # 3 =>  9:11 => (3,4), (3,6)
            0, 1,    # 3 => 11:13 => (3,1), (3,0)

            1, 0, 1, # 4 => 13:16 => (4,5), (4,6), (4,7)
            0, 0, 0, # 4 => 16:19 => (4,2), (4,1), (4,3)

                     # 5 => 19:19 =>
            1, 1,    # 5 => 19:21 => (5,2), (5,4)

            1,       # 6 => 21:22 => (6,7)
            0, 1,    # 6 => 22:24 => (6,4), (6,3)

                     # 7 => 24:24 =>
            1, 1,    # 7 => 24:26 => (7,4), (7,6)
        ])

        # Index for boundary_edges or intern_edges
        assert_equal(idx, [
                     # V0    bounds  edges

            0, 1,    # 0 =>  0:2  => (0,1), (0,3)
                     # 0 =>  2:2  =>

            0, 2, 1, # 1 =>  2:5  => (1,3), (1,2), (1,4)
            0,       # 1 =>  5:6  => (1,0)

            2, 3,    # 2 =>  6:8  => (2,4), (2,5)
            2,       # 2 =>  8:9  => (2,1)

            3, 4,    # 3 =>  9:11 => (3,4), (3,6)
            0, 1,    # 3 => 11:13 => (3,1), (3,0)

            5, 4, 6, # 4 => 13:16 => (4,5), (4,6), (4,7)
            2, 1, 3, # 4 => 16:19 => (4,2), (4,1), (4,3)

                     # 5 => 19:19 =>
            3, 5,    # 5 => 19:21 => (5,2), (5,4)

            7,       # 6 => 21:22 => (6,7)
            4, 4,    # 6 => 22:24 => (6,4), (6,3)

                     # 7 => 24:24 =>
            6, 7,    # 7 => 24:26 => (7,4), (7,6)
        ])

    def assert_memview_equal(self, arr0, arr1):
        """Assert memory view are equal"""
        assert_equal(np.asarray(arr0), np.asarray(arr1))

    def test_neighbours(self):
        edge_map = self.intern_edges.edge_map

        self.assert_memview_equal(edge_map.neighbours(0), [1, 3])
        self.assert_memview_equal(edge_map.neighbours(1), [3, 2, 4, 0])

    def test_label(self):
        label = self.boundary_edges.label

        self.assertEqual(label.shape, (step.NB,))
        self.assertEqual(label[0], 1) # (0,1)
        self.assertEqual(label[1], 2) # (3,0)
        self.assertEqual(label[2], 1) # (1,2)
        self.assertEqual(label[3], 2) # (2,5)
        self.assertEqual(label[4], 2) # (6,3)
        self.assertEqual(label[5], 2) # (5,4)
        self.assertEqual(label[6], 2) # (4,7)
        self.assertEqual(label[7], 3) # (7,6)

    def test_wrong_number_of_label(self):
        label_wrong_number = step.boundary_edge_label[:-1]
        msg = "7 label are given, but there are 8 boundary edges"
        with self.assertRaisesRegex(ValueError, msg):
            self.boundary_edges.add_label(label_wrong_number)

    def test_missing_label(self):
        label_missing = step.boundary_edge_label.copy()
        label_missing[0] = label_missing[-1]
        msg = "Missing label for edge \(0, 1\)"
        with self.assertRaisesRegex(ValueError, msg):
            self.boundary_edges.add_label(label_missing)

    def test_duplicated_label(self):
        label_duplicated = step.boundary_edge_label.copy()
        label_duplicated[1] = label_duplicated[0]
        msg = "Reference for edge \(0, 1\) is given 2 times"
        with self.assertRaisesRegex(ValueError, msg):
            self.boundary_edges.add_label(label_duplicated)

    def test_length(self):
        length = np.asarray(self.boundary_edges.length)
        self.assertEqual(length.shape, (step.NB,))
        assert_equal(length[0], 1)   # (0,1)
        assert_equal(length[1], 1)   # (3,0)
        assert_equal(length[2], 1.5) # (1,2)
        assert_equal(length[3], 1)   # (2,5)
        assert_equal(length[4], 1)   # (6,3)
        assert_equal(length[5], 1.5) # (5,4)
        assert_equal(length[6], 1)   # (4,7)
        assert_equal(length[7], 1)   # (7,6)

    def test_normal(self):
        normal = np.asarray(self.boundary_edges.normal)
        self.assertEqual(normal.shape, (step.NB, 2))
        assert_equal(normal[0], ( 0, -1))  # (0,1)
        assert_equal(normal[1], (-1,  0))  # (3,0)
        assert_equal(normal[2], ( 0, -1))  # (1,2)
        assert_equal(normal[3], ( 1,  0))  # (2,5)
        assert_equal(normal[4], (-1,  0))  # (6,3)
        assert_equal(normal[5], ( 0,  1))  # (5,4)
        assert_equal(normal[6], ( 1,  0))  # (4,7)
        assert_equal(normal[7], ( 0,  1))  # (7,6)

    def test_reorder_intern_edges(self):

        intern_edges = self.intern_edges
        edge_map = intern_edges.edge_map

        # Before reordering:
        #
        # | -------- | --------- |
        # | vertices | triangles |
        # | -------- | --------- |
        # | 1, 3     | 0, 2      |
        # | 1, 4     | 2, 1      |
        # | 2, 4     | 1, 3      |
        # | 3, 4     | 4, 2      |
        # | 4, 6     | 4, 5      |

        # After reordering:
        vertices = np.array([
            [3, 4],
            [3, 1],
            [2, 4],
            [1, 4],
            [6, 4],
        ], dtype='int32')

        intern_edges.reorder(vertices)

        self.assert_memview_equal(edge_map.idx,
                                  [0, 1, 1, 2, 3, 0, 2, 3, 2, 0, 4, 1, 1, 5,
                                   4, 6, 2, 3, 0, 3, 5, 7, 4, 4, 6, 7])
        self.assert_memview_equal(intern_edges.triangles,
                                  [[4, 2], [2, 0], [1, 3], [2, 1], [5, 4]])
        self.assert_memview_equal(intern_edges.vertices, vertices)

    def test_reorder_intern_edges_using_boundary_vertices(self):

        intern_edges = self.intern_edges

        vertices = np.array([
            [2, 5],
            [3, 1],
            [2, 4],
            [1, 4],
            [6, 4],
        ], dtype='int32')

        with self.assertRaisesRegex(ValueError, 'not found'):
            intern_edges.reorder(vertices)

    def test_reorder_intern_edges_using_no_such_vertices(self):

        intern_edges = self.intern_edges

        vertices = np.array([
            [10, 20],
            [3, 1],
            [2, 4],
            [1, 4],
            [6, 4],
        ], dtype='int32')

        with self.assertRaisesRegex(ValueError, 'not found'):
            intern_edges.reorder(vertices)

    def test_reorder_intern_edges_using_duplicated_vertices(self):

        intern_edges = self.intern_edges

        # (6, 4) is duplicated.
        vertices = np.array([
            [3, 4],
            [6, 4],
            [2, 4],
            [1, 4],
            [6, 4],
        ], dtype='int32')

        with self.assertRaisesRegex(ValueError, 'duplicated'):
            intern_edges.reorder(vertices)

    def test_reorder_boundary_edges(self):

        boundary_edges = self.boundary_edges
        edge_map = boundary_edges.edge_map

        # Before reordering:
        #
        # | ----- | -------- | -------- |
        # | index | vertices | triangle |
        # | ----- | -------- | -------- |
        # | 0     | 0, 1     | 0        |
        # | 1     | 3, 0     | 0        |
        # | 2     | 1, 2     | 1        |
        # | 3     | 2, 5     | 3        |
        # | 4     | 6, 3     | 4        |
        # | 5     | 5, 4     | 3        |
        # | 6     | 4, 7     | 5        |
        # | 7     | 7, 6     | 5        |

        # After reordering:
        vertices = np.array([
            [5, 4], # 5  0
            [0, 1], # 0  1
            [1, 2], # 2==2
            [7, 6], # 7  3
            [3, 0], # 1  4
            [6, 3], # 4  5
            [4, 7], # 6==6
            [2, 5], # 3  7
        ], dtype='int32')

        boundary_edges.reorder(vertices)

        self.assert_memview_equal(edge_map.idx,
                                  [1, 4, 0, 2, 1, 1, 2, 7, 2, 3, 5, 0, 4, 0,
                                   4, 6, 2, 1, 3, 7, 0, 3, 4, 5, 6, 3])
        self.assert_memview_equal(boundary_edges.triangle,
                                  [3, 0, 1, 5, 0, 4, 5, 3])
        self.assert_memview_equal(boundary_edges.vertices, vertices)

        boundary_edges.build_next_boundary_edge()

        self.assert_memview_equal(boundary_edges.next_boundary_edge,
                             [6, 2, 7, 5, 1, 4, 3, 0])

    def test_reorder_boundary_edges_using_intern_vertices(self):

        boundary_edges = self.boundary_edges

        vertices = np.array([
            [3, 4], # 5 intern edge
            [0, 1], # 0
            [1, 2], # 2 *
            [7, 6], # 7
            [3, 0], # 1
            [6, 3], # 4
            [4, 7], # 6 *
            [2, 5], # 3
        ], dtype='int32')

        with self.assertRaisesRegex(ValueError, 'not found'):
            boundary_edges.reorder(vertices)

    def test_reorder_boundary_edges_using_undirect_vertices(self):

        boundary_edges = self.boundary_edges

        # After reordering:
        vertices = np.array([
            [4, 5], # 5 not direct
            [0, 1], # 0
            [1, 2], # 2 *
            [7, 6], # 7
            [3, 0], # 1
            [6, 3], # 4
            [4, 7], # 6 *
            [2, 5], # 3
        ], dtype='int32')

        with self.assertRaisesRegex(ValueError, 'not direct'):
            boundary_edges.reorder(vertices)

    def test_reorder_boundary_edges_using_no_such_vertices(self):

        boundary_edges = self.boundary_edges

        vertices = np.array([
            [10, 20], # 5 no such
            [0, 1], # 0
            [1, 2], # 2 *
            [7, 6], # 7
            [3, 0], # 1
            [6, 3], # 4
            [4, 7], # 6 *
            [2, 5], # 3
        ], dtype='int32')

        with self.assertRaisesRegex(ValueError, 'not found'):
            boundary_edges.reorder(vertices)

    def test_reorder_boundary_edges_using_duplicated_vertices(self):

        boundary_edges = self.boundary_edges

        # (6, 4) is duplicated.
        vertices = np.array([
            [5, 4], # 5
            [5, 4], # 0 duplicated
            [1, 2], # 2 *
            [7, 6], # 7
            [3, 0], # 1
            [6, 3], # 4
            [4, 7], # 6 *
            [2, 5], # 3
        ], dtype='int32')

        with self.assertRaisesRegex(ValueError, 'duplicated'):
            boundary_edges.reorder(vertices)

class TestHole(unittest.TestCase):

    def setUp(self):
        self.intern_edges, self.boundary_edges, _ = build_edges(hole.trivtx,
                                                                hole.NV)
        self.edge_map = self.intern_edges.edge_map

    def assert_intern_triangles(self, A, B, T, U):
        """Check that intern edge (A, B) has triangles (T, U)"""
        t, u = self.intern_edges[(A,B)]
        if (t, u) != (T, U):
            raise AssertionError(
                "Expected intern edge ({}, {}) to have triangles ({}, {}),"
                "but has triangles: ({}, {})" .format(T, U, t, u))

    def test_internal_edges(self):
        self.assertEqual(self.intern_edges.vertices.shape, (hole.NI, 2))
        self.assertEqual(self.intern_edges.triangles.shape, (hole.NI, 2))
        self.assert_intern_triangles( 9, 10,  8, 18)
        self.assert_intern_triangles(24, 30, 36, 37)
        self.assert_intern_triangles( 8, 15,  6, 23)

    def test_no_such_intern_edge(self):
        with self.assertRaisesRegex(KeyError, "No such intern edge"):
            self.intern_edges[(10, 18)]

    def assert_boundary_triangle(self, A, B, T):
        """Check that boundary edge (A, B) has triangles T"""
        t = self.boundary_edges[(A,B)]
        if t != T:
            raise AssertionError(
                "Expected boundary edge ({}, {}) to have triangle {},"
                "but has triangle: {}" .format(T, t))

    def test_boundary_edges(self):
        self.assertEqual(self.boundary_edges.vertices.shape, (hole.NB, 2))
        self.assertEqual(self.boundary_edges.triangle.shape, (hole.NB, ))
        self.assert_boundary_triangle( 2 , 3,  2)
        self.assert_boundary_triangle(18, 17, 25)
        self.assert_boundary_triangle(28, 21, 32)

    def test_label(self):
        self.boundary_edges.add_label(hole.boundary_edge_label)

        label = self.boundary_edges.label

        self.assertEqual(label.shape, (hole.NB,))

if __name__ == '__main__':
    unittest.main()



import unittest

from geomalgo import Point3D, Vector3D, Plane

class TestPlane(unittest.TestCase):

    def test_creation(self):

        A = Point3D(0, 0, 0)
        B = Point3D(1, 0, 0)
        C = Point3D(0, 1, 0)

        P = Point3D(0.5, 0.5, 0)
        M = Point3D(0.5, 0.5, 1)

        normal = M - P

        plane = Plane(point=P, normal=normal)

        self.assertEqual(plane.point.x, 0.5)
        self.assertEqual(plane.normal.z, 1)

    def test_point_projection_on_horizontal_plane(self):
        """
                       M(0.5,0.5,1.)
                       +
                       |              ^ z
                       |
        x    B(1,0,0)  |              A(0,0,0)
        <       +------|--------------+
              /        |            /
            /          +          /
          /     P(0.5,0.5,0)    /
        +---------------------+  C(0,1,0)

                          y
        """

        A = Point3D(0, 0, 0)
        B = Point3D(1, 0, 0)
        C = Point3D(0, 1, 0)

        P = Point3D(0.5, 0.5, 0)
        M = Point3D(0.5, 0.5, 1)

        plane = Plane(point=P, normal=M-P)

        projection = plane.project_point(M)
        self.assertAlmostEqual(projection.x, P.x)
        self.assertAlmostEqual(projection.y, P.y)
        self.assertAlmostEqual(projection.z, P.z)

    def test_point_projection_on_vertical_plane(self):
        """
                                     ^ z

                                     + C(0,0,1)
                                   / |
                                 /   |
                               /     |
        M(0.5,0.5,0.5)  x    +   P   |
        x---------------<----|---+   + A(0,0,0)
                             |     /
                             |   /
                             | /
                             +  B(0,1,0)

                         y
        """

        A = Point3D(0,0,0)
        B = Point3D(0,1,0)
        C = Point3D(0,0,1)

        P = Point3D(0. , 0.5, 0.5)
        M = Point3D(0.5, 0.5, 0.5)

        plane = Plane(point=P, normal=M-P)

        projection = plane.project_point(M)
        self.assertAlmostEqual(projection.x, P.x)
        self.assertAlmostEqual(projection.y, P.y)
        self.assertAlmostEqual(projection.z, P.z)

if __name__ == '__main__':
    unittest.main()



from .point3d cimport CPoint3D, Point3D

cdef struct CTriangle3D:
    CPoint3D* A
    CPoint3D* B
    CPoint3D* C

cdef CTriangle3D* new_triangle3d()

cdef void del_triangle3d(CTriangle3D* T)

cdef double compute_area3d(CTriangle3D* T)

cdef void compute_symetric_point3d(CPoint3D* S, CTriangle3D* T, CPoint3D* P)

cdef class Triangle3D:

    cdef public:
        Point3D A
        Point3D B
        Point3D C
        int index



import unittest
from math import pi

from geomalgo import CylindricalPoint

class TestCylindricalPoint(unittest.TestCase):

    def test_to_cartesian(self):
        P = CylindricalPoint(r=2., theta=0., z=0.5)
        A = P.to_cartesian()
        self.assertAlmostEqual(A.x, 2.)
        self.assertAlmostEqual(A.y, 0.)
        self.assertAlmostEqual(A.z, 0.5)

        P = CylindricalPoint(r=2., theta=pi/2, z=0.5)
        A = P.to_cartesian()
        self.assertAlmostEqual(A.x, 0.)
        self.assertAlmostEqual(A.y, 2.)
        self.assertAlmostEqual(A.z, 0.5)

if __name__ == '__main__':
    unittest.main()



from libc.math cimport floor

from ..base2d cimport CPoint2D

cpdef inline int coord_to_index(val, minval, delta):
    """ Point coordinate to box index, in a given coordinate """
    return <int> floor( (val-minval) / delta)


cpdef inline int compute_index(int nx, int ix, int iy):
    return iy*nx + ix


cdef class Grid2D:
    cdef public:
        double xmin
        double xmax
        int nx
        double dx
        double[:] x

        double ymin
        double ymax
        int ny
        double dy
        double[:] y

    cdef void c_find_cell(Grid2D self, Cell2D cell, CPoint2D* P)


cdef class Cell2D:
    cdef public:
        int index

        int ix
        int iy

        double xmin
        double xmax

        double ymin
        double ymax



from geomalgo.base1d cimport *
from geomalgo.base2d cimport *
from geomalgo.base3d cimport *
from geomalgo.grid2d cimport *
from geomalgo.inclusion cimport *
from geomalgo.intersection cimport *
from geomalgo.polar cimport *
from geomalgo.triangulation cimport *



import unittest

import numpy as np
from numpy.testing import assert_equal

import geomalgo as ga

HOLE = ga.data.hole


class TestTriangleToCell(unittest.TestCase):

    def test_four_cells(self):
        """
            25  +-----+-----+-----+-----+
                |     |     |     |     |
                |     |   B |     |     |
                |     |  /|\|     |     |
                |8    | / | \     |   11|
            20  +-----+/--|-+\----+-----+
                |     /   | | \   |     |
                |    /|  1|0|  \  |     |
                |   C-----E-----A |     |   Triangles    Center
                |4   \|  2|3|  /  |    7|   0 ABE        (12, 19)
            15  +-----\---|-+-/---+-----+   1 BCE        ( 8, 19)
                |     |\  | |/    |     |   2 CDE        ( 8, 15)
                |     | \ | /     |     |   3 DAE        (12, 15)
                |     |  \|/|     |     |
                |0    |   D |     |    3|
            10  +-----+-----+-----+-----+
                0     6    12    18    24

        """

        A = ga.Point2D(16, 17, 0)
        B = ga.Point2D(10, 23, 1)
        C = ga.Point2D( 4, 17, 2)
        D = ga.Point2D(10, 11, 3)
        E = ga.Point2D(10, 17, 4)

        points = [A, B, C, D, E]
        x = np.asarray([P.x for P in points])
        y = np.asarray([P.y for P in points])

        triangles = [(A,B,E), (B,C,E), (C,D,E), (D,A,E)]
        trivtx = np.asarray([[P.index for P in vertices]
                            for vertices in triangles], dtype='int32')

        TG = ga.Triangulation2D(x, y, trivtx)
        grid = ga.Grid2D(0, 24, 4, 10, 25, 3)

        bounds = np.zeros((4,4), dtype='int32')

        ga.build_triangle_to_cell(bounds, TG, grid, 0.01)

        assert_equal(bounds, [[1, 3, 1, 3],
                              [0, 2, 1, 3],
                              [0, 2, 0, 2],
                              [1, 3, 0, 2]])


class TestBuildCellToTriangle(unittest.TestCase):

    def test_four_cells(self):
        """
        same geometry as TestTriangleToCell.test_four_cells
        """

        bounds = np.array([[1, 3, 1, 3],
                           [0, 2, 1, 3],
                           [0, 2, 0, 2],
                           [1, 3, 0, 2]], dtype='int32')

        nx, ny = 4, 3

        celltri, celltri_idx = ga.build_cell_to_triangle(bounds, nx, ny)

        assert_equal(celltri, [
            2,           # cell  0     0: 1
            2, 3,        # cell  1     1: 3
            3,           # cell  2     3: 4
                         # cell  3     4: 4
            1, 2,        # cell  4     4: 6
            0, 1, 2, 3,  # cell  5     6:10
            0, 3,        # cell  6    10:12
                         # cell  7    12:12
            1,           # cell  8    12:13
            0, 1,        # cell  9    13:15
            0,           # cell 10    15:16
                         # cell 11    16:16
        ])

        assert_equal(celltri_idx,
                     [0, 1, 3, 4, 4, 6, 10, 12, 12, 13, 15, 16, 16])


class TestTriangulationLocator(unittest.TestCase):

    def test_tri_center(self):
        """
        Check that center of triangles are found in their triangles
        """

        # Triangle centers.
        xcenter = np.average(HOLE.x[HOLE.trivtx], axis=1)
        ycenter = np.average(HOLE.y[HOLE.trivtx], axis=1)
        NP = len(xcenter)

        grid = ga.triangulation.build_grid(HOLE.triangulation, 5, 5)
        locator = ga.TriangulationLocator(HOLE.triangulation, grid)

        triangles = locator.search_points(xcenter, ycenter)
        assert_equal(triangles, np.arange(NP))

        # same test with pre-allocated result.
        locator.search_points(xcenter, ycenter, triangles=np.empty(NP, dtype='int32'))
        assert_equal(triangles, np.arange(NP))

    @unittest.skip
    def test_hole_aligned(self):
        """
        The grid is aligned on triangle vertical and horizontal edges.

        Note: This is articial, because normally grid distance to
        triangulation bounding box is 2*edge_width.

        In particular, we test that the two hole cells have the triangle
        associated, so a point on hole edge is detected.

                  y
                  ^
                 16 +-----+-----+-----+-----+-----+-----+-----+-----+
                    |                                               |
              6     |                                               |
                 15 +    28----------29----------30----------31     +
                    |     |\         /|\         /|\         /|     |
              5     |     | \  33   / | \  36   / | \  39   / |     |
                 14 +     |  \     /  |  \     /  |  \     /  |     +
                    |     |32 \   / 34|35 \   / 37|38 \   / 40|     |
              4     |     |    \ /    |    \ /    |    \ /    |     |
                 13 +    21----22----23----24----25----26----27     +
                    |     |28 / |29 / |           | \ 30| \ 31|     |
              3     |     | / 12| / 13|           | 14\ | 15\ |     |
                 12 +    14----15----16----17----18----19----20     +
                    |     |22 / |23 / |24 / | \ 25| \ 26| \ 27|     |
              2     |     | / 6 | / 7 | / 8 |9  \ |10 \ |11 \ |     |
                 11 +     7-----8-----9----10----11----12----13     +
                    |     |16 / |17 / |18 / | \ 19| \ 20| \ 21|     |
              1     |     | / 0 | / 1 | / 2 |3  \ |4  \ |5  \ |     |
                 10 +     0-----1-----2-----3-----4-----5-----6     +
                    |                                               |
              0     |                                               |
                  9 +-----+-----+-----+-----+-----+-----+-----+-----+
                    -1     0     1     2     3     4     5     6     7   -> x
                        0     1     2     3     4     5     6     7      -> grid.ix
        """
        nx, ny = 8, 7
        dist = 1.
        grid = ga.triangulation.build_grid(HOLE.triangulation, nx, ny, dist)
        locator = ga.TriangulationLocator(HOLE.triangulation, grid)

        # left hole cell
        # triangle 33 is included because its cell range is searched in
        # square of points [21, 23, 29, 28]. An optimisation would be to
        # detect it does not overlap with cell (3, 3).
        self.assertEqual(locator.cell_to_triangles(3, 3),
                         {9, 25, 8, 24, 7, 23, 13, 29, 34, 35, 36, 37, 33})

        # right hole cell
        # triangle 39 is included for the same reason of triangle 33 above.
        self.assertEqual(locator.cell_to_triangles(4, 3),
                         {8, 24, 9, 25, 10, 26, 14, 30, 38, 37, 36, 35, 39})

    def test_points_out(self):
        """
                                      E

                15  28----------29----------30----------31
                     |\         /|\         /|\         /|
                     | \  33   / | \  36   / | \  39   / |
                14   |  \     /  |  \     /  |  \     /  |
                     |32 \   / 34|35 \   / 37|38 \   / 40|
                     |    \ /    |    \ /    |    \ /    |
                13  21----22----23----24----25----26----27
                     |28 / |29 / | A     B   | \ 30| \ 31|
                     | / 12| / 13|           | 14\ | 15\ |
           F    12  14----15----16----17----18----19----20    D
                     |22 / |23 / |24 / | \ 25| \ 26| \ 27|
                     | / 6 | / 7 | / 8 |9  \ |10 \ |11 \ |
                11   7-----8-----9----10----11----12----13
                     |16 / |17 / |18 / | \ 19| \ 20| \ 21|
                     | / 0 | / 1 | / 2 |3  \ |4  \ |5  \ |
                10   0-----1-----2-----3-----4-----5-----6

                     0     1     2     3     4     5     6

                                       C
        """

        # Points out of the domain
        A = ga.Point2D( 2.5, 12.5)
        B = ga.Point2D( 3.5, 12.5)
        C = ga.Point2D( 3  ,  9  )
        D = ga.Point2D(-1  , 12  )
        E = ga.Point2D( 3  , 16  )
        F = ga.Point2D( 7  , 12  )

        points = [A, B, C, D, E, F]
        NP = len(points)

        x = np.array([P.x for P in points])
        y = np.array([P.x for P in points])

        locator = ga.TriangulationLocator(HOLE.triangulation)

        triangles = locator.search_points(x, y)

        assert_equal(triangles, np.full(NP, fill_value=-1, dtype='int32'))

if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from geomalgo import EdgeToTriangles
from geomalgo.data import step, hole

class TestStep(unittest.TestCase):

    def test_normal(self):
        edge2tri = EdgeToTriangles(step.trivtx, step.NV)

        self.assertEqual(edge2tri.NB, step.NB)
        self.assertEqual(edge2tri.NI, step.NI)
        self.assertEqual(edge2tri.NE, step.NE)

        expected = {
            (0, 1): (0, None),
            (0, 3): (0, None),
            (1, 2): (1, None),
            (1, 3): (0, 2),
            (1, 4): (1, 2),
            (2, 4): (1, 3),
            (2, 5): (3, None),
            (3, 4): (2, 4),
            (3, 6): (4, None),
            (4, 5): (3, None),
            (4, 7): (5, None),
            (4, 6): (4, 5),
            (6, 7): (5, None),
        }

        for V0V1, (T0, T1) in expected.items():
            self.assertEqual(edge2tri[V0V1], (T0, T1))

    def test_reversed_key(self):
        edge2tri = EdgeToTriangles(step.trivtx, step.NV)
        self.assertEqual(edge2tri[(4,2)], (1, 3))

    def test_wrong_key(self):
        edge2tri = EdgeToTriangles(step.trivtx, step.NV)
        with self.assertRaisesRegex(KeyError, "No such edge"):
            edge2tri[(3,7)]

    def test_superior_neighbours(self):
        edge2tri = EdgeToTriangles(step.trivtx, step.NV)

        self.assertEqual(edge2tri.superior_neighbours(0), [1, 3])
        self.assertEqual(edge2tri.superior_neighbours(1), [2, 3, 4])
        self.assertEqual(edge2tri.superior_neighbours(2), [4, 5])
        self.assertEqual(edge2tri.superior_neighbours(3), [4, 6])
        self.assertEqual(edge2tri.superior_neighbours(4), [5, 6, 7])
        self.assertEqual(edge2tri.superior_neighbours(5), [])
        self.assertEqual(edge2tri.superior_neighbours(6), [7])
        self.assertEqual(edge2tri.superior_neighbours(7), [])

    def test_inferior_neighbours(self):
        edge2tri = EdgeToTriangles(step.trivtx, step.NV)

        self.assertEqual(edge2tri.inferior_neighbours(0), [])
        self.assertEqual(edge2tri.inferior_neighbours(1), [0])
        self.assertEqual(edge2tri.inferior_neighbours(2), [1])
        self.assertEqual(edge2tri.inferior_neighbours(3), [0, 1])
        self.assertEqual(edge2tri.inferior_neighbours(4), [1, 2, 3])
        self.assertEqual(edge2tri.inferior_neighbours(5), [2, 4])
        self.assertEqual(edge2tri.inferior_neighbours(6), [3, 4])
        self.assertEqual(edge2tri.inferior_neighbours(7), [4, 6])

class TestHole(unittest.TestCase):

    def test_normal(self):
        edge2tri = EdgeToTriangles(hole.trivtx, hole.NV)

        self.assertEqual(edge2tri.NB, hole.NB)
        self.assertEqual(edge2tri.NI, hole.NI)
        self.assertEqual(edge2tri.NE, hole.NE)

if __name__ == '__main__':
    unittest.main()



import unittest

from geomalgo import Vector3D

class TestVector(unittest.TestCase):

    def test_dot_product(self):
        A = Vector3D(1,-2,3)
        B = Vector3D(4,6,5)
        x = A.dot(B)
        self.assertEqual(x, 7)

    def test_norm(self):
        A = Vector3D(1, 0, 0)
        self.assertEqual(A.norm, 1.)

        A = Vector3D(1, 2, 3)
        self.assertAlmostEqual(A.norm, 3.7416573867739413)

    def test_cross_product(self):
        A = Vector3D(1,2,3)
        B = Vector3D(4,6,5)
        V = A * B
        self.assertEqual(V.x, -8)
        self.assertEqual(V.y,  7)
        self.assertEqual(V.z, -2)

    def test_subtract_vector(self):
        A = Vector3D(1,2,3)
        B = Vector3D(4,6,5)
        V = B - A
        self.assertEqual(V.x, 3)
        self.assertEqual(V.y, 4)
        self.assertEqual(V.z, 2)

    def test_normalize(self):
        A = Vector3D(0, 3, 4)
        A.normalize()
        self.assertAlmostEqual(A.x, 0. )
        self.assertAlmostEqual(A.y, 0.6)
        self.assertAlmostEqual(A.z, 0.8)
        self.assertAlmostEqual(A.norm, 1.)

if __name__ == '__main__':
    unittest.main()



__version__ = '0.3.0'

from .base1d import *
from .base2d import *
from .base3d import *
from .cylindrical import *
from .grid2d import *
from .intersection import *
from .polar import *
from .triangulation import *
from . import data



from .edge_map cimport EdgeMap

cdef class BoundaryEdges:
    cdef public:
        int size
        int[:,:] vertices
        int[:] triangle
        int[:] next_boundary_edge
        int[:] label
        double[:] length
        double[:,:] normal
        EdgeMap edge_map



import unittest

from geomalgo import Point3D, Triangle3D

class TestTriangle(unittest.TestCase):

    def test_create_triangle(self):
        A = Point3D(0,0,0)
        B = Point3D(0,1,0)
        C = Point3D(1,0,0)

        triangle = Triangle3D(A,B,C)
        self.assertEqual(triangle.B.y, 1)

        triangle.B.y = 3
        self.assertEqual(triangle.B.y, 3)

        triangle.B = Point3D(-1, -2, -3)
        self.assertEqual(triangle.B.y, -2)

    def test_area(self):
        A = Point3D(0,0,2)
        B = Point3D(0,1,2)
        C = Point3D(1,0,2)

        triangle = Triangle3D(A,B,C)
        self.assertEqual(triangle.area, 0.5)

    def test_symetric_point(self):
        A = Point3D(0,0,0)
        B = Point3D(0,1,0)
        C = Point3D(1,0,0)
        triangle = Triangle3D(A,B,C)

        P = Point3D(1,2,-3)

        S = triangle.symetric_point(P)

        self.assertEqual(S.x, 1)
        self.assertEqual(S.y, 2)
        self.assertEqual(S.z, 3)

if __name__ == '__main__':
    unittest.main()



from ..base3d.point3d cimport CPoint3D
from ..base3d.segment3d cimport CSegment3D
from ..base3d.triangle3d cimport CTriangle3D

cdef int c_intersec3d_triangle_segment(CPoint3D* I, CTriangle3D* tri,
                                       CSegment3D* seg)



import unittest
from math import sqrt, pi

from geomalgo import Point3D

class TestPoint(unittest.TestCase):

    def test_property_x(self):
        A = Point3D(1,2,3)
        self.assertEqual(A.x, 1)
        A.x = 10
        self.assertEqual(A.x, 10)

    def test_index(self):
        A = Point3D(1,2,3)
        self.assertEqual(A.index, 0)

        B = Point3D(1,2,3, index=8)
        self.assertEqual(B.index, 8)

    def test_distance(self):
        A = Point3D(3,2,1)
        B = Point3D(4,3,2)
        dist = A.distance(B)
        expected_dist = sqrt(3.)
        self.assertAlmostEqual(dist, expected_dist)

    def test_vector_from_point_sub_point(self):
        A = Point3D(1,2,3)
        B = Point3D(6,5,4)
        V = B - A
        self.assertEqual(V.x, 5)
        self.assertEqual(V.y, 3)
        self.assertEqual(V.z, 1)

    def test_str(self):
        A = Point3D(3, 2, 1)
        expected_string = "Point3D(3.0, 2.0, 1.0)"
        string = str(A)
        self.assertEqual(string, expected_string)

    def test_to_cylindrical(self):
        A = Point3D(2, 0, 0.5)
        P = A.to_cylindrical()
        self.assertAlmostEqual(P.r, 2)
        self.assertAlmostEqual(P.theta, 0)
        self.assertEqual(P.z, 0.5)

        A = Point3D(0, 2, 0.5)
        P = A.to_cylindrical()
        self.assertAlmostEqual(P.r, 2.)
        self.assertAlmostEqual(P.theta, pi/2)
        self.assertEqual(P.z, 0.5)

if __name__ == '__main__':
    unittest.main()



from .triangulation.triangulation2d cimport (
    Triangulation2D
)

from .triangulation.edge_map cimport EdgeMap

from .triangulation.boundary_edges cimport BoundaryEdges

from .triangulation.intern_edges cimport InternEdges

from .triangulation.interpolator cimport TriangulationInterpolator

from .triangulation.constant cimport INTERN_EDGE, BOUNDARY_EDGE



import unittest
from math import sqrt, pi

from geomalgo import Point2D, Vector2D

class TestPoint2D(unittest.TestCase):

    def test_property_x(self):
        A = Point2D(1,2)
        self.assertEqual(A.x, 1)
        A.x = 10
        self.assertEqual(A.x, 10)

    def test_index(self):
        A = Point2D(1,2)
        self.assertEqual(A.index, 0)

        B = Point2D(1,2, index=8)
        self.assertEqual(B.index, 8)

    def test_distance(self):
        A = Point2D(2,1)
        B = Point2D(3,2)
        dist = A.distance(B)
        expected_dist = sqrt(2.)
        self.assertAlmostEqual(dist, expected_dist)

    def test_add_vector(self):
        A = Point2D(1, 2)
        v = Vector2D(3, 1)
        B = A + v
        self.assertEqual(B.x, 4)
        self.assertEqual(B.y, 3)

    def test_vector_from_point_sub_point(self):
        A = Point2D(1,2)
        B = Point2D(4,3)
        V = B - A
        self.assertEqual(V.x, 3)
        self.assertEqual(V.y, 1)

    def test_str(self):
        A = Point2D(2,1)
        expected_string = "Point2D(2.0, 1.0)"
        string = str(A)
        self.assertEqual(string, expected_string)

    def test_to_polar(self):
        A = Point2D(2, 0)
        P = A.to_polar()
        self.assertAlmostEqual(P.r, 2)
        self.assertAlmostEqual(P.theta, 0)

        A = Point2D(0, 2)
        P = A.to_polar()
        self.assertAlmostEqual(P.r, 2.)
        self.assertAlmostEqual(P.theta, pi/2)

class TestIsLeft(unittest.TestCase):

    def test_is_left(self):
        """
        P

        A----B
        """
        A = Point2D(0, 0)
        B = Point2D(1, 0)
        P = Point2D(0, 1)
        self.assertTrue(P.is_left(A,B))

    def test_is_right(self):
        """
        A----B

        P
        """
        A = Point2D(0,  0)
        B = Point2D(1,  0)
        P = Point2D(0, -1)
        self.assertFalse(P.is_left(A,B))

    def test_on_line(self):
        """
        A----B----P
        """
        A = Point2D(0, 0)
        B = Point2D(1, 0)
        P = Point2D(2, 0)
        with self.assertRaisesRegex(ValueError, "Point is on line \(AB\)"):
            P.is_left(A,B)


class TestEquality(unittest.TestCase):

    def test_equal(self):
        A = Point2D(2, 1)
        B = Point2D(2, 1)
        self.assertEqual(A, B)

    def test_not_equal(self):
        A = Point2D(1, 1)
        B = Point2D(2, 1)
        C = Point2D(1, 2)
        self.assertNotEqual(A, B)
        self.assertNotEqual(A, C)


if __name__ == '__main__':
    unittest.main()



import unittest

from geomalgo import symetric1d

class TestSymetric1D(unittest.TestCase):

    def test_normal(self):
        x = 1
        center = 3
        self.assertEqual(symetric1d(center,x), 5)

if __name__ == '__main__':
    unittest.main()



from ..base2d cimport CTriangle2D

cdef class Triangulation2D:

    cdef public:
        int NV
        int NT
        double[:] x
        double[:] y
        int[:,:] trivtx

    cdef void get(Triangulation2D self, int I, CTriangle2D* T)



from libc.math cimport fabs

from .point2d cimport CPoint2D, Point2D, is_left


# ============================================================================
# Structures
# ============================================================================


cdef struct CTriangle2D:
    CPoint2D* A
    CPoint2D* B
    CPoint2D* C


cdef CTriangle2D* new_triangle2d()


cdef void del_triangle2d(CTriangle2D* ctri2d)


cdef inline void triangle2d_set(CTriangle2D* ABC,
                                CPoint2D* A, CPoint2D* B, CPoint2D* C):
    ABC.A = A
    ABC.B = B
    ABC.C = C


# ============================================================================
# Computational functions
# ============================================================================


cdef bint triangle2d_includes_point2d(CTriangle2D* ABC, CPoint2D* P,
                                      double edge_width_square)


cdef int triangle2d_on_edges(CTriangle2D* ABC, CPoint2D* P,
                             double edge_width_square)


cdef inline double triangle2d_signed_area(CTriangle2D* T):
    return 0.5 * is_left(T.A, T.B, T.C)


cdef inline double triangle2d_area(CTriangle2D* T):
    return fabs(0.5 * is_left(T.A, T.B, T.C))


cdef inline void triangle2d_center(CTriangle2D* T, CPoint2D* C):
    C.x = (T.A.x + T.B.x + T.C.x) / 3.
    C.y = (T.A.y + T.B.y + T.C.y) / 3.


cdef void triangle2d_gradx_grady_det(CTriangle2D* T, double signed_area,
                                     double gradx[3], double grady[3],
                                     double det[3])


# ============================================================================
# Python API
# ============================================================================


cdef class Triangle2D:
    cdef public:
        int index
        double signed_area
        double gradx[3]
        double grady[3]
        double det[3]

    cdef:
        Point2D A
        Point2D B
        Point2D C
        CTriangle2D ctri2d

    cdef _set_precomputed(Triangle2D self, Point2D A, Point2D B, Point2D C,
                          int index, double signed_area, double gradx[3],
                          double grady[3], double det[3])

    cdef alloc_new(Triangle2D self)



from libc.math cimport sqrt

from .vector2d cimport CVector2D, Vector2D


# ============================================================================
# Structures
# ============================================================================


cdef struct CPoint2D:
    double x
    double y


cdef CPoint2D* new_point2d()


cdef void del_point2d(CPoint2D* cpoint2d)


cdef inline bint point2d_equal(CPoint2D* A, CPoint2D* B):
    return A.x == B.x and A.y == B.y


# ============================================================================
# Computational functions
# ============================================================================


cdef void subtract_points2d(CVector2D* AB, const CPoint2D* B,
                            const CPoint2D* A)

cdef void point2d_plus_vector2d(CPoint2D* result, CPoint2D* start,
                                double factor, CVector2D* vector)

cdef inline double is_left(CPoint2D* A, CPoint2D* B, CPoint2D* P):
    return (B.x - A.x) * (P.y - A.y) \
         - (P.x - A.x) * (B.y - A.y)

cdef inline double point2d_distance(CPoint2D* A, CPoint2D* B):
    return sqrt(point2d_square_distance(A, B))

cdef inline double point2d_square_distance(CPoint2D* A, CPoint2D* B):
    return (B.x-A.x)**2 + (B.y-A.y)**2


# ============================================================================
# Python API
# ============================================================================


cdef class Point2D:
    cdef public:
        int index
        """Optional point index, for example fo a triangulation."""

        str name
        """Optional point name, to annotate a plot with."""
    cdef:
        CPoint2D* cpoint2d



import unittest

import geomalgo as ga


STEP = ga.data.step


class TestTriangulation(unittest.TestCase):

    def test_get(self):
        TG = ga.Triangulation2D(STEP.x, STEP.y, STEP.trivtx)

        triangle = TG[3]

        self.assertEqual(triangle.index, 3)

        self.assertEqual(triangle.A.x, 2.5)
        self.assertEqual(triangle.A.y,  10)

        self.assertEqual(triangle.B.x, 2.5)
        self.assertEqual(triangle.B.y,  11)

        self.assertEqual(triangle.C.x, 1)
        self.assertEqual(triangle.C.y, 11)

        self.assertAlmostEqual(triangle.area, 0.75)

    def test_to_numpy(self):
        TG = ga.Triangulation2D(STEP.x, STEP.y, STEP.trivtx)
        x, y, trivtx = TG.to_numpy()

    def test_to_matplotlib(self):
        TG = ga.Triangulation2D(STEP.x, STEP.y, STEP.trivtx)
        tri = TG.to_matplotlib()

if __name__ == '__main__':
    unittest.main()



import unittest
from math import sqrt

import numpy as np

from geomalgo import Point2D, Segment2D, Segment2DCollection

class TestSegment2D(unittest.TestCase):

    def test_create_segment(self):
        A = Point2D(1,2)
        B = Point2D(3,4)

        # ======================
        # Create segment2d.
        # ======================
        segment = Segment2D(A,B)
        self.assertEqual(segment.B.y, 4)

        self.assertEqual(segment.u.x, 2)
        self.assertEqual(segment.u.y, 2)
        self.assertAlmostEqual(segment.length, 2*sqrt(2))

        M = segment.compute_middle()
        self.assertAlmostEqual(M.x, 2)
        self.assertAlmostEqual(M.y, 3)

        # ======================
        # Modify B.y
        # ======================
        segment.B.y = 5
        self.assertEqual(segment.B.y, 5)

        segment.recompute()
        self.assertEqual(segment.u.x, 2)
        self.assertEqual(segment.u.y, 3)
        self.assertAlmostEqual(segment.length, sqrt(13))

        M = segment.compute_middle()
        self.assertAlmostEqual(M.x, 2)
        self.assertAlmostEqual(M.y, 3.5)

        # ======================
        # Modify B
        # ======================
        segment.B = Point2D(-1, -2)
        self.assertEqual(segment.B.y, -2)

        segment.recompute()
        self.assertEqual(segment.u.x, -2)
        self.assertEqual(segment.u.y, -4)
        self.assertAlmostEqual(segment.length, 2*sqrt(5))

        M = segment.compute_middle()
        self.assertAlmostEqual(M.x, 0)
        self.assertAlmostEqual(M.y, 0)

    def test_from_tuple(self):
        segment = Segment2D((1, 2), (3, 4))
        self.assertEqual(segment.A.x, 1)
        self.assertEqual(segment.A.y, 2)

        self.assertEqual(segment.B.x, 3)
        self.assertEqual(segment.B.y, 4)

class TestIncludesPoint(unittest.TestCase):

    def test_horizontal_segment(self):
        """

        O      A      I      B      P
        +------+------+------+------+
        -0.5   0      0.5    1      1.5
        """

        A = Point2D(0, 0)
        B = Point2D(1, 0)
        segment = Segment2D(A,B)

        O = Point2D(-0.5, 0)
        I = Point2D( 0.5, 0)
        P = Point2D( 1.5, 0)

        self.assertTrue( segment.includes_collinear_point(I) )

        self.assertTrue( segment.includes_collinear_point(A) )
        self.assertTrue( segment.includes_collinear_point(B) )

        self.assertFalse( segment.includes_collinear_point(O) )
        self.assertFalse( segment.includes_collinear_point(P) )

    def test_vertical_segment(self):
        """
         1.5  + P
              |
         1    + B
              |
         0.5  + I
              |
         0    + A
              |
        -0.5  + 0
        """

        A = Point2D(0, 0)
        B = Point2D(0, 1)
        segment = Segment2D(A,B)

        O = Point2D(0, -0.5)
        I = Point2D(0,  0.5)
        P = Point2D(0,  1.5)

        self.assertTrue( segment.includes_collinear_point(I) )

        self.assertTrue( segment.includes_collinear_point(A) )
        self.assertTrue( segment.includes_collinear_point(B) )

        self.assertFalse( segment.includes_collinear_point(O) )
        self.assertFalse( segment.includes_collinear_point(P) )


class TestSegment2DAt(unittest.TestCase):

    def test_horizontal_segment(self):
        """
        A-----P-----+-----+-----B
        0     0.5   1     1.5   2.
        """

        A = Point2D(0, 3)
        B = Point2D(2, 3)
        segment = Segment2D(A, B)

        P = segment.at(0.25)

        self.assertAlmostEqual(P.x, 0.5)
        self.assertAlmostEqual(P.y, 3)

    def test_vertical_segment(self):
        """
          2  B
             |
        1.5  +
             |
          1  +
             |
        0.5  P
             |
          0  A
        """

        A = Point2D(3, 0)
        B = Point2D(3, 2)
        segment = Segment2D(A, B)

        P = segment.at(0.25)

        self.assertAlmostEqual(P.x, 3)
        self.assertAlmostEqual(P.y, 0.5)

    def test_diagonal_segment(self):
        """
            2                  B
          1.5              +
            1          +
          0.5      P
            0  A
               4   6   8   10  12
        """

        A = Point2D( 4, 0)
        B = Point2D(12, 2)
        segment = Segment2D(A, B)

        P = segment.at(0.25)

        self.assertAlmostEqual(P.x, 6)
        self.assertAlmostEqual(P.y, 0.5)

class TestSegment2DWhere(unittest.TestCase):

    def test_horizontal_segment(self):
        """
        A-----P-----+-----+-----B
        0     0.5   1     1.5   2.
        """

        A = Point2D(0, 3)
        B = Point2D(2, 3)
        P = Point2D(0.5, 3)
        segment = Segment2D(A, B)

        coord = segment.where(P)

        self.assertAlmostEqual(coord, 0.25)

    def test_vertical_segment(self):
        """
          2  B
             |
        1.5  +
             |
          1  +
             |
        0.5  P
             |
          0  A
        """

        A = Point2D(3, 0)
        B = Point2D(3, 2)
        P = Point2D(3, 0.5)
        segment = Segment2D(A, B)

        coord = segment.where(P)

        self.assertAlmostEqual(coord, 0.25)

    def test_diagonal_segment(self):
        """
            2                  B
          1.5              +
            1          +
          0.5      P
            0  A
               4   6   8   10  12
        """

        A = Point2D( 4, 0)
        B = Point2D(12, 2)
        P = Point2D(6, 0.25)
        segment = Segment2D(A, B)

        coord = segment.where(P)

        self.assertAlmostEqual(coord, 0.25)

class TestSegment2DCollection(unittest.TestCase):

    def test_get(self):
        """
        12  E------F

        11  C------D

        10  A------B
            1      2
        """

        xa = np.array([1, 1, 1], dtype='d')
        xb = np.array([2, 2, 2], dtype='d')

        ya = np.array([10, 11, 12], dtype='d')
        yb = np.array([10, 11, 12], dtype='d')

        collection = Segment2DCollection(xa, xb, ya, yb)
        CD = collection[1]

        self.assertEqual(CD.A.x,  1)
        self.assertEqual(CD.A.y, 11)

        self.assertEqual(CD.B.x,  2)
        self.assertEqual(CD.B.y, 11)

        self.assertAlmostEqual(CD.length, 1)


class TestPoint2dDistance(unittest.TestCase):

    def test_horizontal_line(self):
        """
              P

        A-----------B
        """

        A = Point2D(1, 2)
        B = Point2D(3, 2)
        AB = Segment2D(A, B)

        P = Point2D(2, 5)
        self.assertEqual(AB.point_distance(P), 3)

        P = Point2D(-3, 5)
        self.assertEqual(AB.point_distance(P), 5)

        P = Point2D(7, 5)
        self.assertEqual(AB.point_distance(P), 5)

    def test_vertical_line(self):
        """
        B
        |
        |  B
        |
        A
        """

        A = Point2D(1, 2)
        B = Point2D(1, 4)
        AB = Segment2D(A, B)

        P = Point2D(4, 3)
        self.assertEqual(AB.point_distance(P), 3)

        P = Point2D(4, -2)
        self.assertEqual(AB.point_distance(P), 5)

        P = Point2D(4, 8)
        self.assertEqual(AB.point_distance(P), 5)


    def test_on_horizontal_line(self):
        """
        A-----P-----B
        """

        A = Point2D(1, 2)
        B = Point2D(3, 2)
        AB = Segment2D(A, B)

        P = Point2D(2, 2)
        self.assertEqual(AB.point_distance(P), 0)

        P = Point2D(-3, 2)
        self.assertEqual(AB.point_distance(P), 4)

        P = Point2D(7, 2)
        self.assertEqual(AB.point_distance(P), 4)

    def test_on_vertical_line(self):
        """
        B
        |
        B
        |
        A
        """

        A = Point2D(1, 2)
        B = Point2D(1, 4)
        AB = Segment2D(A, B)

        P = Point2D(1, 3)
        self.assertEqual(AB.point_distance(P), 0)

        P = Point2D(1, -2)
        self.assertEqual(AB.point_distance(P), 4)

        P = Point2D(1, 8)
        self.assertEqual(AB.point_distance(P), 4)

if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from geomalgo import Point2D, Triangle2D


class TestTriangle2D(unittest.TestCase):
    """
      1 C
        | \
        |   \
        |     \
      0 A-------B
        0       1
    """

    def test_create(self):
        A = Point2D(0,0)
        B = Point2D(1,0)
        C = Point2D(0,1)

        triangle = Triangle2D(A, B, C, index=4)
        self.assertAlmostEqual(triangle.area, 0.5)

        # Change coordiante, and recompute
        triangle.A.y = -1
        triangle.recompute()
        self.assertAlmostEqual(triangle.area, 1)

        # Change point, and recompute
        triangle.A = Point2D(0,0)
        triangle.recompute()
        self.assertAlmostEqual(triangle.area, 0.5)

    def test_degenerated(self):
        """
        A----B----P
        """
        A = Point2D(0,0)
        B = Point2D(1,0)
        C = Point2D(2,0)

        with self.assertRaisesRegex(ValueError, "Triangle is degenerated"):
            triangle = Triangle2D(A, B, C)


class TestAreaAndCounterclockwise(unittest.TestCase):

    def test_counterclockwise(self):
        """
        C
        +
        |  \
        |     \
        +--------+
        A        B
        """
        A = Point2D(0,0)
        B = Point2D(1,0)
        C = Point2D(0,1)

        triangle = Triangle2D(A, B, C)

        self.assertAlmostEqual(triangle.signed_area, 0.5)
        self.assertAlmostEqual(triangle.area, 0.5)
        self.assertTrue(triangle.counterclockwise)

        triangle = Triangle2D(A, B, C, force_counterclockwise=True)

        self.assertAlmostEqual(triangle.signed_area, 0.5)
        self.assertAlmostEqual(triangle.area, 0.5)
        self.assertTrue(triangle.counterclockwise)

    def test_clockwise(self):
        """
        B
        +
        |  \
        |     \
        +--------+
        A        C
        """
        A = Point2D(0,0)
        B = Point2D(0,1)
        C = Point2D(1,0)

        triangle = Triangle2D(A, B, C)

        self.assertAlmostEqual(triangle.signed_area, -0.5)
        self.assertAlmostEqual(triangle.area, 0.5)
        self.assertFalse(triangle.counterclockwise)

        triangle = Triangle2D(A, B, C, force_counterclockwise=True)

        self.assertAlmostEqual(triangle.signed_area, 0.5)
        self.assertAlmostEqual(triangle.area, 0.5)
        self.assertTrue(triangle.counterclockwise)


class TestCenter(unittest.TestCase):

    def test_normal(self):

        A = Point2D(0,0)
        B = Point2D(3,0)
        C = Point2D(0,6)

        triangle = Triangle2D(A, B, C)

        C = triangle.center

        self.assertAlmostEqual(C.x, 1.)
        self.assertAlmostEqual(C.y, 2.)


class TestInterpolate(unittest.TestCase):

    def f(self, x, y):
        return 3*x - 4*y + 1

    def check_interpolate_at(self, x, y):
        P = Point2D(x, y)
        actual = self.triangle.interpolate(self.data, P)
        expected = self.f(x, y)
        self.assertAlmostEqual(actual, expected)

    def test_normal(self):
        """
          C
        1 +
          |  \
          |     \
        0 +--------+
          A        B
          0        1
        """

        A = Point2D(0, 0)
        B = Point2D(1, 0)
        C = Point2D(0, 1)
        self.triangle = Triangle2D(A, B, C)

        x = np.array( [A.x, B.x, C.x] )
        y = np.array( [A.y, B.y, C.y] )
        self.data = self.f(x, y)

        self.check_interpolate_at(A.x, A.y)
        self.check_interpolate_at(B.x, B.y)
        self.check_interpolate_at(C.x, C.y)
        self.check_interpolate_at(0.25, 0.25)
        self.check_interpolate_at(0.1, 0.2)
        self.check_interpolate_at(0.2, 0.1)

    def test_clockwise(self):

        """
          C
        1 +
          |  \
          |     \
        0 +--------+
          A        B
          0        1
        """

        A = Point2D(0, 0)
        C = Point2D(1, 0)
        B = Point2D(0, 1)
        self.triangle = Triangle2D(A, B, C)

        x = np.array( [A.x, B.x, C.x] )
        y = np.array( [A.y, B.y, C.y] )
        self.data = self.f(x, y)

        self.check_interpolate_at(A.x, A.y)
        self.check_interpolate_at(B.x, B.y)
        self.check_interpolate_at(C.x, C.y)
        self.check_interpolate_at(0.25, 0.25)
        self.check_interpolate_at(0.1, 0.2)
        self.check_interpolate_at(0.2, 0.1)

if __name__ == '__main__':
    unittest.main()



from .point3d cimport CPoint3D, Point3D

cdef struct CSegment3D:
    CPoint3D* A
    CPoint3D* B

cdef CSegment3D* new_segment3d()

cdef void del_segment3d(CSegment3D* csegment3d)

cdef class Segment3D:

    cdef public:
        Point3D A
        Point3D B



from .base3d.plane cimport (
    CPlane, new_plane, del_plane, point_projection_to_plane, Plane
)

from .base3d.point3d cimport (
    CPoint3D, new_point3d, del_point3d, subtract_points3d,
    point3d_plus_vector3d, c_point3d_distance, Point3D
)

from .base3d.segment3d cimport (
    CSegment3D, new_segment3d, del_segment3d, Segment3D
)

from .base3d.triangle3d cimport (
    Triangle3D, CTriangle3D, new_triangle3d, del_triangle3d, compute_area3d,
    compute_symetric_point3d, Triangle3D
)

from .base3d.vector3d cimport (
    CVector3D, new_vector3d, del_vector3d, cross_product3d, subtract_vector3d,
    dot_product3d, compute_norm3d, normalize_vector3d, Vector3D
)



from libc.math cimport sqrt

from .vector3d cimport CVector3D, Vector3D

cdef struct CPoint3D:
    double x
    double y
    double z

cdef CPoint3D* new_point3d()

cdef void del_point3d(CPoint3D* cpoint3d)

cdef void subtract_points3d(CVector3D * u, const CPoint3D * B,
                            const CPoint3D * A)

cdef void point3d_plus_vector3d(CPoint3D* result, CPoint3D* start,
                                double factor, CVector3D* vector)

cdef inline double c_point3d_distance(CPoint3D* A, CPoint3D* B):
    return sqrt((B.x-A.x)**2 + (B.y-A.y)**2 + (B.z-A.z)**2)

cdef class Point3D:
    cdef public:
        int index
    cdef:
        CPoint3D* cpoint3d



import unittest
from math import pi

from geomalgo import PolarPoint

class TestPolarPoint(unittest.TestCase):

    def test_to_cartesian(self):
        P = PolarPoint(r=2., theta=0.)
        A = P.to_cartesian()
        self.assertAlmostEqual(A.x, 2.)
        self.assertAlmostEqual(A.y, 0.)

        P = PolarPoint(r=2., theta=pi/2)
        A = P.to_cartesian()
        self.assertAlmostEqual(A.x, 0.)
        self.assertAlmostEqual(A.y, 2.)

if __name__ == '__main__':
    unittest.main()



import unittest

from geomalgo import Point2D, Segment2D

class TestSegment(unittest.TestCase):

    def test_intersection(self):
        """
        4        S
                 |
        3  P-----I--Q
                 |
        1        |
                 |
        0        R
           0  1  2  3
        """

        # CASE12
        P,Q = Point2D(0, 3), Point2D(3, 3)
        R,S = Point2D(2, 0), Point2D(2, 4)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertAlmostEqual(I0.x, 2.)
        self.assertAlmostEqual(I0.y, 3.)
        self.assertIsNone(I1)

        self.assertAlmostEqual(coords[0], 2./3.)
        self.assertAlmostEqual(coords[1], 3./4.)
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_parallel(self):
        """
        4

        3  P--------Q

        1  R--------S

        0
           0  1  2  3

        """

        #CASE00
        P,Q = Point2D(0, 3), Point2D(3, 3)
        R,S = Point2D(0, 1), Point2D(3, 1)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertIsNone(I0)
        self.assertIsNone(I1)

        self.assertIsNone(coords[0])
        self.assertIsNone(coords[1])
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_equal(self):
        """
        4

        3  P--------Q
           R        S
        1

        0
           0  1  2  3
        """

        P,Q = Point2D(0, 3), Point2D(3, 3)
        R,S = Point2D(0, 3), Point2D(3, 3)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertAlmostEqual(I0.x, P.x)
        self.assertAlmostEqual(I0.y, P.y)

        self.assertAlmostEqual(I1.x, Q.x)
        self.assertAlmostEqual(I1.y, Q.y)

        self.assertEqual(len(coords), 4)
        self.assertAlmostEqual(coords[0], 0)
        self.assertAlmostEqual(coords[1], 0)
        self.assertAlmostEqual(coords[2], 1)
        self.assertAlmostEqual(coords[3], 1)

    def test_colinear_no_overlap(self):
        """

        3  P-----Q     R-----S

           0     1     2     3
        """

        # CASE07
        P,Q = Point2D(0, 3), Point2D(1, 3)
        R,S = Point2D(2, 3), Point2D(3, 3)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertIsNone(I0)
        self.assertIsNone(I1)

        self.assertIsNone(coords[0])
        self.assertIsNone(coords[1])
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_colinear_intersect_one_point(self):
        """
                Q R
        3  P-----+-----S

           0     1     2
        """

        # CASE08
        P,Q = Point2D(0, 3), Point2D(1, 3)
        R,S = Point2D(1, 3), Point2D(2, 3)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertAlmostEqual(I0.x, 1)
        self.assertAlmostEqual(I0.y, 3)
        self.assertIsNone(I1)

        self.assertAlmostEqual(coords[0], 1)
        self.assertAlmostEqual(coords[1], 0)
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_colinear_overlap_segment(self):
        """
        3  P-----+-----Q-----+
                 R           S
           0   1   2   3
        """

        # CASE09
        P,Q = Point2D(0,   3), Point2D(3, 3)
        R,S = Point2D(1.5, 3), Point2D(4.5, 3)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertAlmostEqual(I0.x, R.x)
        self.assertAlmostEqual(I0.y, R.y)

        self.assertAlmostEqual(I1.x, Q.x)
        self.assertAlmostEqual(I1.y, Q.y)

        self.assertAlmostEqual(coords[0], 0.5)
        self.assertAlmostEqual(coords[1], 0)
        self.assertAlmostEqual(coords[2], 1)
        self.assertAlmostEqual(coords[3], 0.5)

    def test_no_intersection(self):
        """
        4              S
                       |
        3  P--------Q  |
                       |
        1              |
                       |
        0              R
           0  1  2  3  4
        """

        # CASE10
        P,Q = Point2D(0, 3), Point2D(3, 3)
        R,S = Point2D(4, 0), Point2D(4, 4)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertIsNone(I0)
        self.assertIsNone(I1)

        self.assertIsNone(coords[0])
        self.assertIsNone(coords[1])
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

        # CASE11
        I0, I1, coords = RS.intersect_segment(PQ, return_coords=True)

        self.assertIsNone(I0)
        self.assertIsNone(I1)

        self.assertIsNone(coords[0])
        self.assertIsNone(coords[1])
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_extremity(self):
        """
        4
                    S
        3  P--------Q
                    |
        1           |
                    |
        0           R
           0  1  2  3
        """

        P,Q = Point2D(0, 3), Point2D(3, 3)
        R,S = Point2D(3, 0), Point2D(3, 3)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertAlmostEqual(I0.x, 3.)
        self.assertAlmostEqual(I0.y, 3.)
        self.assertIsNone(I1)

        self.assertAlmostEqual(coords[0], 1)
        self.assertAlmostEqual(coords[1], 1)
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_same_point(self):
        """
            P Q
          1  +
            R S
             1
        """

        # CASE02
        P,Q = Point2D(1, 1), Point2D(1, 1)
        R,S = Point2D(1, 1), Point2D(1, 1)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertAlmostEqual(I0.x, 1)
        self.assertAlmostEqual(I0.y, 1)
        self.assertIsNone(I1)

        self.assertAlmostEqual(coords[0], 0)
        self.assertAlmostEqual(coords[1], 0)
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_distinct_point(self):
        """
            P Q     R S
           1 +       +
             1       2
        """

        # CASE01
        P,Q = Point2D(1, 1), Point2D(1, 1)
        R,S = Point2D(1, 2), Point2D(1, 2)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertIsNone(I0)
        self.assertIsNone(I1)

        self.assertIsNone(coords[0])
        self.assertIsNone(coords[1])
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_collinear_point_not_in_segment(self):
        """
             P Q
           1  +   R-----S
              1   2     3
        """

        # CASE03
        P,Q = Point2D(1, 1), Point2D(1, 1)
        R,S = Point2D(1, 2), Point2D(1, 3)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertIsNone(I0)
        self.assertIsNone(I1)

        self.assertIsNone(coords[0])
        self.assertIsNone(coords[1])
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

        # CASE05
        I0, I1, coords = RS.intersect_segment(PQ, return_coords=True)

        self.assertIsNone(I0)
        self.assertIsNone(I1)

        self.assertIsNone(coords[0])
        self.assertIsNone(coords[1])
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

    def test_collinear_point_in_segment(self):
        """
               P Q
           1 R--+--S
             1  2  3
        """

        # CASE04
        P,Q = Point2D(2, 1), Point2D(2, 1)
        R,S = Point2D(1, 1), Point2D(3, 1)

        PQ = Segment2D(P, Q)
        RS = Segment2D(R, S)

        I0, I1, coords = PQ.intersect_segment(RS, return_coords=True)

        self.assertAlmostEqual(I0.x, 2)
        self.assertAlmostEqual(I0.y, 1)
        self.assertIsNone(I1)

        self.assertAlmostEqual(coords[0], 0)
        self.assertAlmostEqual(coords[1], 0.5)
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

        # CASE06
        I0, I1, coords = RS.intersect_segment(PQ, return_coords=True)

        self.assertAlmostEqual(I0.x, 2)
        self.assertAlmostEqual(I0.y, 1)
        self.assertIsNone(I1)

        self.assertAlmostEqual(coords[0], 0.5)
        self.assertAlmostEqual(coords[1], 0)
        self.assertIsNone(coords[2])
        self.assertIsNone(coords[3])

if __name__ == '__main__':
    unittest.main()



# An edge (V0,V1), with V1>V0 is stored in an array at index V0. (V0,V1) may
# belong to 1 triangle (if it is a boundary edge), or 2 triangles (if it is an
# internal edge). `next_edge` points to the next edge connected to V0, or has
# value NULL.
cdef struct Edge:
    int V1
    bint has_two_triangles
    int T0
    int T1
    # If True, (T0, V0, V1) is counterclockwise and (T1, V0, V1) is clockwise.
    # If False, (T0, V0, V1) is clockwise and (T1, V0, V1) is
    # counterclockwise.
    bint counterclockwise
    Edge* next_edge

# Edges (V0,V1) with V1<V0. This store duplicated edge, as (V0,V1) is stored
# in Edge, and (V1,V0) is InferiorEdge. Given a vertice V0, this allow to
# find all V1 vertice connected by edge with 0(1) performance.
cdef struct InferiorEdge:
    int V0
    InferiorEdge* next_inferior_edge

# Edge** is an array of linked list of `Edge*`, indexed with V0, with V1>V0.
# `edges_number` counts the total number of `Edges` in each linked list.
cdef struct CEdgeToTriangles:
    # Number of vertices
    int NV
    # Number of edges (NI+NB)
    int NE
    # Number of internal edges.
    int NI
    # Number of boundary edges
    int NB
    Edge** edges
    InferiorEdge** inferior_edges

# Allocate and initialize a new `Edge`.
cdef Edge* edge_new(int V1, int T0, bint counterclockwise)

# Allocate and initialize a new `InferiorEdge`.
cdef InferiorEdge* inferior_edge_new(int V1)

# Allocate and initialize a new edge_to_triangles_new.
cdef CEdgeToTriangles* edge_to_triangles_new(int NV)

# Delete a `CEdgeToTriangles`.
cdef void edge_to_triangle_del(CEdgeToTriangles* edge_to_triangles)

# Add a new (V0,V1) -> T entry in the mapping.
cdef void edge_to_triangles_add(CEdgeToTriangles* edge_to_triangles,
                                int V0, int V1, int T)

# Compute all (V0,V1) -> (T0, T1) mappings
cdef void edge_to_triangles_compute(CEdgeToTriangles* edge_to_triangles,
                                    int[:,:] trivtx)

# Retrieve a (V0,V1) -> (T0, T1) entry.
cdef Edge* edge_to_triangles_get(CEdgeToTriangles* edge_to_triangles,
                                       int V0, int V1)

cdef void edge_to_triangles_display(CEdgeToTriangles* edge2tri)

cdef class EdgeToTriangles:
    cdef:
        CEdgeToTriangles* _edge2tri
    cdef public:
        int NE
        int NI
        int NB



from libc.math cimport sqrt


# ============================================================================
# Structures
# ============================================================================


cdef struct CVector2D:
    double x
    double y

cdef CVector2D* new_vector2d()

cdef void del_vector2d(CVector2D* V)


# ===========================================================================
# Computational functions
# ============================================================================


cdef inline void vector2d_times_scalar(CVector2D *t, double alpha, CVector2D *u):
    t.x = alpha * u.x
    t.y = alpha * u.y


cdef inline void subtract_vector2d(CVector2D *AB, CVector2D *AC, CVector2D *BC):
    AB.x = AC.x - BC.x
    AB.y = AC.y - BC.y


cdef inline void add_vector2d(CVector2D *AC, CVector2D *AB, CVector2D *BC):
    AC.x = AB.x + BC.x
    AC.y = AB.y + BC.y


cdef inline double cross_product2d(CVector2D *u, CVector2D *v):
    return u.x*v.y - u.y*v.x


cdef inline double dot_product2d(CVector2D *u, CVector2D *v):
    return u.x*v.x + u.y*v.y


cdef inline double compute_norm2d(CVector2D *u):
    return sqrt(u.x*u.x + u.y*u.y)


cdef inline void normalize_vector2d(CVector2D *u):
    cdef:
        double norm = compute_norm2d(u)
    u.x /= norm
    u.y /= norm


cdef inline void compute_normal2d(CVector2D *n, CVector2D *u, double norm):
    n.x =  u.y / norm
    n.y = -u.x / norm


# ============================================================================
# Python API
# ============================================================================


cdef class Vector2D:
    cdef:
        CVector2D* cvector2d

